@@ -33,8 +33,12 @@
 #include <glib/gi18n.h>
 #include <gio/gio.h>
 
+#define LIBSOUP_USE_UNSTABLE_REQUEST_API
 #include <libsoup/soup-gnome.h>
+#include <libsoup/soup-requester.h>
+
 #include "gvfsbackendhttp.h"
+#include "gvfshttpinputstream.h"
 #include "gvfsjobopenforread.h"
 #include "gvfsjobread.h"
 #include "gvfsjobseekread.h"
@@ -49,9 +53,6 @@
 #include "gvfsdaemonprotocol.h"
 #include "gvfsdaemonutils.h"
 
-#include "soup-input-stream.h"
-
-
 G_DEFINE_TYPE (GVfsBackendHttp, g_vfs_backend_http, G_VFS_TYPE_BACKEND)
 
 static void
@@ -94,6 +95,8 @@ g_vfs_backend_http_init (GVfsBackendHttp *backend)
   backend->session_async = soup_session_async_new_with_options ("user-agent",
                                                                 "gvfs/" VERSION,
                                                                 NULL);
+  /* SoupRequester seems to depend on use-thread-context */
+  g_object_set (G_OBJECT (backend->session_async), "use-thread-context", TRUE, NULL);
 
   /* Proxy handling */
   proxy_resolver = g_object_new (SOUP_TYPE_PROXY_RESOLVER_GNOME, NULL);
@@ -118,6 +121,10 @@ g_vfs_backend_http_init (GVfsBackendHttp *backend)
   soup_session_add_feature (backend->session_async, content_decoder);
   g_object_unref (content_decoder);
 
+  /* Request API */
+  soup_session_add_feature_by_type (backend->session, SOUP_TYPE_REQUESTER);
+  soup_session_add_feature_by_type (backend->session_async, SOUP_TYPE_REQUESTER);
+
   /* Logging */
   debug = g_getenv ("GVFS_HTTP_DEBUG");
   if (debug)
@@ -234,25 +241,25 @@ http_error_code_from_status (guint status)
 }
 
 
-static void
-g_vfs_job_failed_from_http_status (GVfsJob *job, guint status_code, const char *message)
+void
+http_job_failed (GVfsJob *job, SoupMessage *msg)
 {
-  switch (status_code) {
+  switch (msg->status_code) {
 
   case SOUP_STATUS_NOT_FOUND:
     g_vfs_job_failed_literal (job, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
-                              message);
+                              msg->reason_phrase);
     break;
 
   case SOUP_STATUS_UNAUTHORIZED:
   case SOUP_STATUS_PAYMENT_REQUIRED:
   case SOUP_STATUS_FORBIDDEN:
     g_vfs_job_failed (job, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
-                      _("HTTP Client Error: %s"), message);
+                      _("HTTP Client Error: %s"), msg->reason_phrase);
     break;
   default:
     g_vfs_job_failed (job, G_IO_ERROR, G_IO_ERROR_FAILED,
-                      _("HTTP Error: %s"), message);
+                      _("HTTP Error: %s"), msg->reason_phrase);
   }
 }
 
@@ -337,6 +344,7 @@ open_for_read_ready (GObject      *source_object,
 {
   GInputStream *stream;
   GVfsJob      *job;
+  SoupMessage  *msg;
   gboolean      res;
   gboolean      can_seek;
   GError       *error;
@@ -345,9 +353,9 @@ open_for_read_ready (GObject      *source_object,
   error  = NULL;
   job    = G_VFS_JOB (user_data);
 
-  res = soup_input_stream_send_finish (stream,
-                                       result,
-                                       &error);
+  res = g_vfs_http_input_stream_send_finish (stream,
+					     result,
+					     &error);
   if (res == FALSE)
     {
       g_vfs_job_failed_literal (G_VFS_JOB (job),
@@ -360,6 +368,16 @@ open_for_read_ready (GObject      *source_object,
       return;
     }
 
+  msg = g_vfs_http_input_stream_get_message (stream);
+  if (!SOUP_STATUS_IS_SUCCESSFUL (msg->status_code))
+    {
+      http_job_failed (G_VFS_JOB (job), msg);
+      g_object_unref (msg);
+      g_object_unref (stream);
+      return;
+    }
+  g_object_unref (msg);
+
   can_seek = G_IS_SEEKABLE (stream) && g_seekable_can_seek (G_SEEKABLE (stream));
 
   g_vfs_job_open_for_read_set_can_seek (G_VFS_JOB_OPEN_FOR_READ (job), can_seek);
@@ -387,22 +405,16 @@ http_backend_open_for_read (GVfsBackend *backend,
 {
   GVfsBackendHttp *op_backend;
   GInputStream    *stream;
-  SoupMessage     *msg;
 
   op_backend = G_VFS_BACKEND_HTTP (backend);
 
-  msg = soup_message_new_from_uri (SOUP_METHOD_GET, uri);
+  stream = g_vfs_http_input_stream_new (op_backend->session_async, uri);
 
-  soup_message_body_set_accumulate (msg->response_body, FALSE);
-
-  stream = soup_input_stream_new (op_backend->session_async, msg);
-  g_object_unref (msg);
-
-  soup_input_stream_send_async (stream,
-                                G_PRIORITY_DEFAULT,
-                                job->cancellable,
-                                open_for_read_ready,
-                                job);
+  g_vfs_http_input_stream_send_async (stream,
+				      G_PRIORITY_DEFAULT,
+				      job->cancellable,
+				      open_for_read_ready,
+				      job);
 }
 
 /* *** read () *** */
@@ -579,15 +591,12 @@ file_info_from_message (SoupMessage *msg,
   /* read http/1.1 rfc, until then we copy the local files
    * behaviour */ 
   if (basename != NULL &&
-      g_file_attribute_matcher_matches (matcher,
-                                        G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME))
+      (g_file_attribute_matcher_matches (matcher,
+                                         G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME) ||
+       g_file_attribute_matcher_matches (matcher,
+                                         G_FILE_ATTRIBUTE_STANDARD_EDIT_NAME)))
     ed_name = gvfs_file_info_populate_names_as_local (info, basename);
 
-  if (ed_name != NULL &&
-      g_file_attribute_matcher_matches (matcher,
-                                        G_FILE_ATTRIBUTE_STANDARD_EDIT_NAME))
-    g_file_info_set_edit_name (info, ed_name);
-
   g_free (basename);
   g_free (ed_name);
 
@@ -655,8 +664,7 @@ query_info_ready (SoupSession *session,
 
   if (! SOUP_STATUS_IS_SUCCESSFUL (msg->status_code))
     {
-      g_vfs_job_failed_from_http_status (G_VFS_JOB (job), msg->status_code,
-                                         msg->reason_phrase);
+      http_job_failed (G_VFS_JOB (job), msg);
       return;
     }
 
@@ -693,7 +701,7 @@ try_query_info_on_read (GVfsBackend           *backend,
                         GFileInfo             *info,
                         GFileAttributeMatcher *attribute_matcher)
 {
-    SoupMessage *msg = soup_input_stream_get_message (G_INPUT_STREAM (handle));
+    SoupMessage *msg = g_vfs_http_input_stream_get_message (G_INPUT_STREAM (handle));
 
     file_info_from_message (msg, info, attribute_matcher);
     g_object_unref (msg);
