@@ -9,43 +9,6 @@
 #include <stdlib.h>
 #include <time.h>
 
-#if HAVE_SYS_STATFS_H
-#include <sys/statfs.h>
-#endif
-#if HAVE_SYS_STATVFS_H
-#include <sys/statvfs.h>
-#endif
-#if HAVE_SYS_VFS_H
-#include <sys/vfs.h>
-#elif HAVE_SYS_MOUNT_H
-#if HAVE_SYS_PARAM_H
-#include <sys/param.h>
-#endif
-#include <sys/mount.h>
-#endif
-
-#if defined(HAVE_STATFS) && defined(HAVE_STATVFS)
-/* Some systems have both statfs and statvfs, pick the
-   most "native" for these */
-# if !defined(HAVE_STRUCT_STATFS_F_BAVAIL)
-   /* on solaris and irix, statfs doesn't even have the
-      f_bavail field */
-#  define USE_STATVFS
-# else
-  /* at least on linux, statfs is the actual syscall */
-#  define USE_STATFS
-# endif
-
-#elif defined(HAVE_STATFS)
-
-# define USE_STATFS
-
-#elif defined(HAVE_STATVFS)
-
-# define USE_STATVFS
-
-#endif
-
 #include "metatree.h"
 #include "metabuilder.h"
 #include <glib.h>
@@ -169,7 +132,8 @@ struct _MetaTree {
   MetaJournal *journal;
 };
 
-static void         meta_tree_refresh_locked   (MetaTree    *tree);
+static void         meta_tree_refresh_locked   (MetaTree    *tree,
+						gboolean     force_reread);
 static MetaJournal *meta_journal_open          (MetaTree    *tree,
 						const char  *filename,
 						gboolean     for_write,
@@ -276,50 +240,6 @@ meta_tree_clear (MetaTree *tree)
     }
 }
 
-static gboolean
-is_on_nfs (char *filename)
-{
-#ifdef USE_STATFS
-  struct statfs statfs_buffer;
-  int statfs_result;
-#elif defined(USE_STATVFS) && defined(HAVE_STRUCT_STATVFS_F_BASETYPE)
-  struct statvfs statfs_buffer;
-  int statfs_result;
-#endif
-  char *dirname;
-  gboolean res;
-
-  dirname = g_path_get_dirname (filename);
-
-  res = FALSE;
-
-#ifdef USE_STATFS
-
-# if STATFS_ARGS == 2
-  statfs_result = statfs (dirname, &statfs_buffer);
-# elif STATFS_ARGS == 4
-  statfs_result = statfs (dirname, &statfs_buffer,
-			  sizeof (statfs_buffer), 0);
-# endif
-  if (statfs_result == 0)
-#ifdef __OpenBSD__
-    res = strcmp(statfs_buffer.f_fstypename, MOUNT_NFS) == 0;
-#else
-    res = statfs_buffer.f_type == 0x6969;
-#endif
-
-#elif defined(USE_STATVFS) && defined(HAVE_STRUCT_STATVFS_F_BASETYPE)
-  statfs_result = statvfs (dirname, &statfs_buffer);
-
-  if (statfs_result == 0)
-    res = strcmp (statfs_buffer.f_basetype, "nfs") == 0;
-#endif
-
-  g_free (dirname);
-
-  return res;
-}
-
 static gboolean
 link_to_tmp (const char *source, char *tmpl)
 {
@@ -430,7 +350,7 @@ meta_tree_init (MetaTree *tree)
 
   retried = FALSE;
  retry:
-  tree->on_nfs = is_on_nfs (tree->filename);
+  tree->on_nfs = meta_builder_is_on_nfs (tree->filename);
   fd = safe_open (tree, tree->filename, O_RDONLY);
   if (fd == -1)
     {
@@ -510,7 +430,7 @@ meta_tree_init (MetaTree *tree)
      journal. However we can detect this case by looking at the tree and see
      if its been rotated, we do this to ensure we have an uptodate tree+journal
      combo. */
-  meta_tree_refresh_locked (tree);
+  meta_tree_refresh_locked (tree, FALSE);
 
   return TRUE;
 
@@ -552,6 +472,12 @@ meta_tree_exists (MetaTree *tree)
   return tree->fd != -1;
 }
 
+gboolean
+meta_tree_is_on_nfs (MetaTree *tree)
+{
+  return tree->on_nfs;
+}
+
 static GHashTable *cached_trees = NULL;
 G_LOCK_DEFINE_STATIC (cached_trees);
 
@@ -658,10 +584,10 @@ meta_tree_has_new_journal_entries (MetaTree *tree)
 
 /* Must be called with a write lock held */
 static void
-meta_tree_refresh_locked (MetaTree *tree)
+meta_tree_refresh_locked (MetaTree *tree, gboolean force_reread)
 {
   /* Needs to recheck since we dropped read lock */
-  if (meta_tree_needs_rereading (tree))
+  if (force_reread || meta_tree_needs_rereading (tree))
     {
       if (tree->header)
 	meta_tree_clear (tree);
@@ -685,7 +611,7 @@ meta_tree_refresh (MetaTree *tree)
   if (needs_refresh)
     {
       g_rw_lock_writer_lock (&metatree_lock);
-      meta_tree_refresh_locked (tree);
+      meta_tree_refresh_locked (tree, FALSE);
       g_rw_lock_writer_unlock (&metatree_lock);
     }
 }
@@ -844,24 +770,6 @@ meta_data_get_key (MetaTree *tree,
   return dataent;
 }
 
-static char *
-get_journal_filename (const char *filename, guint32 random_tag)
-{
-  const char *hexdigits = "0123456789abcdef";
-  char tag[9];
-  int i;
-
-  for (i = 7; i >= 0; i--)
-    {
-      tag[i] = hexdigits[random_tag % 0x10];
-      random_tag >>= 4;
-    }
-
-  tag[8] = 0;
-
-  return g_strconcat (filename, "-", tag, ".log", NULL);
-}
-
 static void
 meta_journal_free (MetaJournal *journal)
 {
@@ -1145,20 +1053,30 @@ meta_journal_open (MetaTree *tree, const char *filename, gboolean for_write, gui
   char *data;
   char *journal_filename;
   int open_flags, mmap_prot;
+  gboolean retried;
 
   g_assert (sizeof (MetaJournalHeader) == 20);
-
-  journal_filename = get_journal_filename (filename, tag);
+  retried = FALSE;
 
   if (for_write)
     open_flags = O_RDWR;
   else
     open_flags = O_RDONLY;
 
+ retry:
+  journal_filename = meta_builder_get_journal_filename (filename, tag);
   fd = safe_open (tree, journal_filename, open_flags);
   g_free (journal_filename);
   if (fd == -1)
-    return NULL;
+    {
+      if (errno == ENOENT && tree->for_write && !retried)
+        {
+          retried = TRUE;
+          if (meta_builder_create_new_journal (filename, tag))
+            goto retry;
+        }
+      return NULL;
+    }
 
   if (fstat (fd, &statbuf) != 0 ||
       statbuf.st_size < sizeof (MetaJournalHeader))
@@ -1303,6 +1221,11 @@ meta_journal_iterate (MetaJournal *journal,
     {
       sizep = (guint32 *)entry;
       entry = (MetaJournalEntry *)((char *)entry - GUINT32_FROM_BE (*(sizep-1)));
+      if (GUINT32_FROM_BE (*(sizep)) < sizeof (MetaJournalEntry) && entry > journal->first_entry)
+        {
+          g_warning ("meta_journal_iterate: found short sized entry, possible journal corruption\n");
+          break;
+        }
 
       mtime = GUINT64_FROM_BE (entry->mtime);
       journal_path = &entry->path[0];
@@ -2342,6 +2265,13 @@ apply_journal_to_builder (MetaTree *tree,
 
       sizep = (guint32 *)entry;
       entry = (MetaJournalEntry *)((char *)entry + GUINT32_FROM_BE (*(sizep)));
+      if (GUINT32_FROM_BE (*(sizep)) < sizeof (MetaJournalEntry) && entry < journal->last_entry)
+        {
+          /* This shouldn't happen, we found an entry that is shorter than its data */
+          /* See https://bugzilla.gnome.org/show_bug.cgi?id=637095 for discussion */
+          g_warning ("apply_journal_to_builder: found short sized entry, possible journal corruption\n");
+          break;
+        }
     }
 }
 
@@ -2363,7 +2293,8 @@ meta_tree_flush_locked (MetaTree *tree)
   res = meta_builder_write (builder,
 			    meta_tree_get_filename (tree));
   if (res)
-    meta_tree_refresh_locked (tree);
+    /* Force re-read since we wrote a new file */
+    meta_tree_refresh_locked (tree, TRUE);
 
   meta_builder_free (builder);
 
